#! /usr/bin/env python3

################################################################################
""" Git module V2 - now implemented as a wrapper around pygit2 (work in progress)

    Copyright (C) 2023 John Skilleter

    Licence: GPL v3 or later

    Except where stated otherwise, functions in this module:

    * Return the output from the equivalent git command as an array of
      strings.

    * Functions will raise exceptions on error. If the underlying git command
      returns an error, a git.GitError() exception is raised.

    * TODO: [ ] Cache list of branches when git.branches/isbranch called
    * TODO: [ ] API change - git_run_status should raise an exception on failure and just return the git output
"""
################################################################################

import os
import shutil
import sys
import re
import logging
import fnmatch
import subprocess

import pygit2

from . import run
from . import gitlab

################################################################################
# Configuration files to access

(WORKTREE, LOCAL, GLOBAL, SYSTEM) = list(range(4))

# Options always passed to Git (disable autocorrect to stop it running the wrong command
# if an invalid command name has been specified).

STANDARD_GIT_OPTIONS = ['-c', 'help.autoCorrect=never']

# Default default branches (can be overridden in .gitconfig via skilleter-thingy.defaultBranches

DEFAULT_DEFAULT_BRANCHES = 'develop,main,master'

################################################################################

class GitError(run.RunError):
    """Run exception."""

    def __init__(self, msg, exit_status=1):
        super().__init__(msg, exit_status)

################################################################################

def git(cmd, stdout=None, stderr=None, path=None):
    """ Wrapper for run.run that raises a GitError instead of RunError
        so that Git module users do not to include the run module just
        to get the exception.
        Optionally redirect stdout and stderr as specified. """

    git_cmd = ['git'] + STANDARD_GIT_OPTIONS

    if path:
        git_cmd += ['-C', path]

    git_cmd += cmd if isinstance(cmd, list) else [cmd]

    logging.debug('Running %s', ' '.join(git_cmd))

    sys.stdout.flush()
    sys.stderr.flush()

    try:
        return run.run(git_cmd, stdout=stdout, stderr=stderr)
    except run.RunError as exc:
        raise GitError(exc.msg, exc.status) from exc
    except FileNotFoundError as exc:
        raise GitError(exc, 1)

################################################################################

def git_run_status(cmd, stdout=None, stderr=None, path=None, redirect=True):
    """ Wrapper for subprocess.run that returns the output and status, and
        does not raise an exception on error.
        Optionally redirect stdout and stderr as specified. """

    logging.debug('Running git %s', ' '.join(cmd))

    git_cmd = ['git'] + STANDARD_GIT_OPTIONS

    if path:
        git_cmd += ['-C', path]

    git_cmd += cmd if isinstance(cmd, list) else [cmd]

    sys.stdout.flush()
    sys.stderr.flush()

    if redirect:
        result = subprocess.run(git_cmd,
                                stdout=stdout or subprocess.PIPE,
                                stderr=stderr or subprocess.PIPE,
                                text=True, check=False,
                                errors='ignore',
                                universal_newlines=True)
    else:
        result = subprocess.run(git_cmd, check=False)

    return (result.stdout or result.stderr), result.returncode

################################################################################

def clone(reponame, working_tree=None):
    """ Clone a repo """

    cmd = ['clone', reponame]

    if working_tree:
        cmd.append(working_tree)

    return git(cmd)

################################################################################

def init(reponame, bare=False, path=None):
    """ Initialise a new working tree """

    cmd = ['init']

    if bare:
        cmd.append('--bare')

    cmd.append(reponame)

    return git(cmd, path=path)

################################################################################

def iscommit(commit, remote=False, remote_only=False, path=None):
    """ Return True if "commit" is a valid SHA1, branch or tag
        If remote==True then if there are no direct matches it will also
        check for a matching remote branch
        If remote_only==True, then ONLY remote branches will be checked."""

    # Fail on empty commit

    if not commit:
        return False

    # Optionally look for an exact matching local branch or commit

    if not remote_only:
        try:
            git(['cat-file', '-t', commit], path=path)
            return True

        except GitError:
            pass

    # Optionally look for matching remote branch

    if remote or remote_only:
        for remote in remote_names(path=path):
            try:
                if git(['ls-remote', remote, commit], path=path):
                    return True

            except GitError:
                pass

    # Nothing found

    return False

################################################################################

def branch(branchname='HEAD', path=None):
    """ Return the name of the current git branch or None"""

    try:
        return git(['symbolic-ref', '--short', '-q', branchname], path=path)[0]
    except GitError:
        return None

################################################################################

def tag(path=None):
    """ If the current commit is tagged, return the tag(s) or None """

    try:
        return git(['describe', '--tags', '--exact-match'], path=path)[0]
    except GitError:
        return None

################################################################################

def tags(path=None):
    """ Return the list of tags in the current repo """

    return git(['tag', '--list'], path=path)

################################################################################

def tag_delete(tag, push=False, path=None):
    """Delete a tag, optionally pushing the deletion"""

    git(['tag', '-d', tag], path=path)

    if push:
        git(['push', 'origin', '--delete', tag], path=path)

################################################################################

def tag_apply(tag, commit=None, push=False, path=None):
    """Apply a tag, optionally pushing it"""

    cmd = ['tag', tag]

    if commit:
        cmd.append(commit)

    git(cmd, path=path)

    if push:
        git(['push', 'origin', tag], path=path)

################################################################################

def current_commit(short=False, path=None):
    """ Return the SHA1 of the current commit """

    cmd = ['rev-parse']

    if short:
        cmd.append('--short')

    cmd.append('HEAD')

    return git(cmd, path=path)[0]

################################################################################

def pull(repo=None, all=False, path=None):
    """ Run a git pull """

    cmd = ['pull']

    if all:
        cmd.append('--all')

    if repo:
        cmd.append(repo)

    return git(cmd, path=path)

################################################################################

def checkout(branch, create=False, commit=None, path=None):
    """ Checkout a branch (optionally creating it or creating it from the
        specified commit) """

    cmd = ['checkout']

    if create or commit:
        cmd.append('-b')

    cmd.append(branch)

    if commit:
        cmd.append(commit)

    return git(cmd, path=path)

################################################################################

def merge(branch, path=None):
    """ Merge a branch """

    cmd = ['merge', branch]

    return git(cmd, path=path)

################################################################################

def abort_merge(path=None):
    """ Abort the current merge """

    return git(['merge', '--abort'], path=path)

################################################################################

def set_upstream(branch, upstream=None, path=None):
    """ Set the default upstream branch """

    if not upstream:
        upstream = f'origin/{branch}'

    cmd = ['branch', f'--set-upstream-to={upstream}', branch]

    return git(cmd, path=path)

################################################################################

def fetch(all=False, path=None):
    """ Run git fetch """

    cmd = ['fetch']

    if all:
        cmd.append('--all')

    return git(cmd, path=path)

################################################################################

def rebase_required(branch, parent, path=None):
    """ Return True if the specified branch needs to be rebased against its
        parent.
    """

    # Find the latest commit on the parent branch and the most recent commit
    # that both branches have in common.

    parent_tip = git(['show-ref', '--heads', '-s', parent], path=path)
    common_commit = git(['merge-base', parent, branch], path=path)

    # Different commits, so rebase is required

    return parent_tip[0] != common_commit[0]

################################################################################

def rebase(branch, path=None):
    """ Rebase the current branch against the specified branch """

    return git_run_status(['rebase', branch], path=path)

################################################################################

def abort_rebase(path=None):
    """ Abort the current rebase """

    return git(['rebase', '--abort'], path=path)

################################################################################

def rebasing(path=None):
    """ Return True if currently rebasing, False otherwise """

    gitdir = git_dir(path=path)

    return os.path.isdir(os.path.join(gitdir, 'rebase-apply')) or \
        os.path.isdir(os.path.join(gitdir, 'rebase-merge'))

################################################################################

def bisecting(path=None):
    """ Return True if currently rebasing, False otherwise """

    gitdir = git_dir(path=path)

    return os.path.isfile(os.path.join(gitdir, 'BISECT_START'))

################################################################################

def merging(path=None):
    """ Return True if currently merging, False otherwise """

    gitdir = git_dir(path=path)

    return os.path.isfile(os.path.join(gitdir, 'MERGE_MODE'))

################################################################################

def remotes(path=None):
    """ Return the list of git remotes """

    repo = pygit2.Repository(git_dir(path=path))

    git_remotes = {}

    for name in repo.remotes.names():
        git_remotes[name] = repo.remotes[name].url

    return git_remotes

################################################################################

def remote_names(path=None):
    """ Return the list of remote names """

    repo = pygit2.Repository(git_dir(path=path))

    results = list(repo.remotes.names())

    return results

################################################################################

def project(short=False, path=None):
    """ Return the name of the current git project """

    git_remotes = remotes(path=path)
    name = ''

    for remote in git_remotes:
        try:
            if '//' in git_remotes[remote]:
                name = git_remotes[remote].split('//')[-1].split('/', 1)[-1]
                break

            if '@' in git_remotes[remote]:
                name = git_remotes[remote].split(':')[-1]
                break

        except ValueError:
            continue

    if name.endswith('.git'):
        name = name[:-4]

    if short:
        name = os.path.basename(name)

    return name

################################################################################

def status_info(ignored=False, untracked=False, path=None, all_untracked=False):
    """ Git status, optionally include files ignored in .gitignore and/or
        untracked files.
        Returns data in the same dictionary format as used by commit_info() """

    cmd = ['status', '-z']

    if ignored:
        cmd.append('--ignored')

    if all_untracked:
        cmd.append('--untracked-files=all')

    results = git(cmd, path=path)

    # Dictionary of results, indexed by filename where the status is 2 characters
    # the first representing the state of the file in the index and the second the state
    # of the file in the working tree where:
    # M=modified, A=added, D=deleted, R=renamed, C=copied, U=unmerged, ?=untracked, !=ignored
    # Where a file has been renamed we don't report the original name

    info = {}

    if results:
        result = results[0].split('\0')

        for r in result:
            if len(r) > 3 and r[2] == ' ':
                git_status = r[0:2]
                name = r[3:]

                info[name] = git_status

    return info

################################################################################

def status(ignored=False, untracked=False, path=None, all_untracked=False):
    """ Git status, optionally include files ignored in .gitignore and/or
        untracked files.
        Similar to status_info, but returns data as a list, rather than a
        dictionary. """

    cmd = ['status', '--porcelain']

    if ignored:
        cmd.append('--ignored')

    if all_untracked:
        cmd.append('--untracked-files=all')

    results = git(cmd, path=path)

    # Nested list of results. For each entry:
    # item 0 is the status where: M=modified, A=added, D=deleted, R=renamed, C=copied, U=unmerged, ?=untracked, !=ignored
    # item 1 is the name
    # item 2 (if present) is the old name in cases where a file has been renamed.

    # TODO: This can't handle the case where a file has ' -> ' in the filename
    # TODO: This can't handle the case where a filename is enclosed in double quotes in the results

    info = []

    for result in results:
        stats = []
        stats.append(result[0:2])

        if not untracked and result[0] == '?':
            continue

        name = result[3:]
        if ' -> ' in name:
            stats += name.split(' -> ', 1)
        else:
            stats.append(name)

        info.append(stats)

    return info

################################################################################

def working_tree(path=None):
    """ Location of the current working tree or None if we are not in a working tree """

    repo_dir = git_dir(path=path)

    if repo_dir:
        return os.path.abspath(os.path.join(repo_dir, os.pardir))

    return None

################################################################################

def git_dir(path=None):
    """ Return the relative path to the .git directory """

    if not path:
        path = os.getcwd()

    return pygit2.discover_repository(path)

################################################################################

def tree_path(filename, path=None):
    """ Normalise a filename (absolute or relative to the current directory)
        so that it is relative to the top-level directory of the working tree """

    git_tree = working_tree(path=path)

    return os.path.relpath(filename, git_tree)

################################################################################

def difftool(commit_1=None, commit_2=None, files=None, tool=None, path=None):
    """ Run git difftool """

    cmd = ['difftool']

    if tool:
        cmd += ['--tool', tool]

    if commit_1:
        cmd.append(commit_1)

    if commit_2:
        cmd.append(commit_2)

    if files:
        cmd.append('--')

        if isinstance(files, str):
            cmd.append(files)
        else:
            cmd += files

    return git(cmd, path=path)

################################################################################

# Match 'git diff --numstat' output - first re splits into lines added, removed
# and name. Second one is used if a file has been renamed, to get the old and
# new name components.

_DIFF_OUTPUT_RE = re.compile(r'(-|\d+)\s+(-|\d+)\s+(.*)')
_DIFF_OUTPUT_RENAME_RE = re.compile(r'(.*)\{(.*) => (.*)\}(.*)')

def commit_info(commit_1=None, commit_2=None, paths=None, diff_stats=False, path=None):
    """ Return details of changes either in single commit (defaulting to the most
        recent one) or between two commits, optionally restricted a path or paths
        and optionally returning diff statistics, with and/or without taking
        whitespace into account.
    """

    def parse_diff_output(result):
        """Extract previous and current filename (which may be the same) and lines added/deleted
           from output from git diff --numstat"""

        p = _DIFF_OUTPUT_RE.fullmatch(result)

        # This shouldn't happen...

        if not p:
            return 'ERROR', 'ERROR', 0, 0

        # Extract number of lines added/removed

        lines_ins = 0 if p.group(1) == '-' else int(p.group(1))
        lines_del = 0 if p.group(2) == '-' else int(p.group(2))

        # Check for rename and get both old and new names
        # The replace of '//' with '/' is to cope with the case where a file is moved
        # to a parent or subdirectory where the git output will be 'PARENT/{SUBDIR => }/NAME'
        # and the new name would otherwise be 'PARENT//NAME'

        if ' => ' in p.group(3):
            q = _DIFF_OUTPUT_RENAME_RE.fullmatch(p.group(3))

            if q:
                old_filename = (q.group(1) + q.group(2) + q.group(4)).replace('//', '/')
                new_filename = (q.group(1) + q.group(3) + q.group(4)).replace('//', '/')
            else:
                old_filename, new_filename = p.group(3).split(' => ')
        else:
            old_filename = new_filename = p.group(3)

        return old_filename, new_filename, lines_ins, lines_del

    # Either get changes between the two commits, or changes in the specified commit

    params = []

    if commit_1:
        params.append(commit_1)

    if commit_2:
        params.append(commit_2)

    if paths:
        params.append('--')

        if isinstance(paths, str):
            params.append(paths)
        else:
            params += paths

    results = git(['diff', '--name-status'] + params, path=path)

    # Parse the output

    info = {}

    for result in results:
        if result == '':
            continue

        # Renames and copies have an extra field (which we ignore) for the destination
        # file. We just get the status (Add, Move, Delete, Type-change, Copy, Rename)
        # and the name.

        if result[0] in ('R', 'C'):
            filestatus, oldname, filename = result.split('\t')
            info[filename] = {'status': filestatus[0], 'oldname': oldname}
        else:
            filestatus, filename = result.split('\t')
            info[filename] = {'status': filestatus[0], 'oldname': filename}

    # Add the diff stats, if requested

    if diff_stats:
        # Run git diff to get stats, and add them to the info
        # TODO: Need to extract old name of renamed files

        results = git(['diff', '--numstat'] + params, path=path)

        for result in results:
            old_filename, new_filename, lines_ins, lines_del = parse_diff_output(result)

            info[new_filename]['deleted'] = lines_del
            info[new_filename]['added'] = lines_ins

        # Run git diff to get stats ignoring whitespace changes and add them

        results = git(['diff', '--numstat', '--ignore-all-space', '--ignore-blank-lines'] + params, path=path)

        for result in results:
            old_filename, new_filename, lines_ins, lines_del = parse_diff_output(result)

            info[new_filename]['non-ws deleted'] = lines_del
            info[new_filename]['non-ws added'] = lines_ins

        # Fill in the blanks - files

        for filename in info:
            if 'deleted' not in info[filename]:
                info[filename]['deleted'] = info[filename]['added'] = 0

            if 'non-ws deleted' not in info[filename]:
                info[filename]['non-ws deleted'] = info[filename]['non-ws added'] = 0

    return info

################################################################################

def diff(commit=None, renames=True, copies=True, relative=False, path=None):
    """ Return a list of differences between two commits, working tree and a commit or working tree and head """

    if commit:
        if isinstance(commit, list):
            if len(commit) > 2:
                raise GitError('git.diff - invalid parameters')

        else:
            commit = [commit]
    else:
        commit = []

    cmd = ['diff', '--name-status']

    if renames:
        cmd.append('--find-renames')

    if copies:
        cmd.append('--find-copies')

    if relative:
        cmd.append('--relative')

    cmd += commit

    return git(cmd, path=path)

################################################################################

def diff_status(commit1, commit2='HEAD', path=None):
    """ Return True if there is no difference between the two commits, False otherwise """

    cmd = ['diff', '--no-patch', '--exit-code', commit1, commit2]

    try:
        git(cmd, path=path)
    except GitError:
        return False

    return True

################################################################################

def show(revision, filename, outfile=None, path=None):
    """ Return the output from git show revision:filename """

    return git(['show', f'{revision}:{filename}'], stdout=outfile, path=path)

################################################################################

def add(files, path=None):
    """ Add file to git """

    return git(['add'] + files, path=path)

################################################################################

def rm(files, path=None):
    """ Remove files from git """

    return git(['rm'] + files, path=path)

################################################################################

def commit(files=None,
           message=None,
           all=False, amend=False, foreground=False, patch=False, dry_run=False,
           path=None):
    """ Commit files to git """

    cmd = ['commit']

    if amend:
        cmd.append('--amend')

    if all:
        cmd.append('--all')

    if patch:
        cmd.append('--patch')
        foreground = True

    if dry_run:
        cmd.append('--dry-run')

    if files is not None:
        cmd += files

    if message:
        cmd += ['-m', message]

    if foreground:
        return git(cmd, stdout=sys.stdout, stderr=sys.stderr, path=path)

    return git(cmd, path=path)

################################################################################

def push(all=False, mirror=False, tags=False, atomic=False, dry_run=False,
         follow_tags=False, receive_pack=False, repo=None, force=False, delete=False,
         prune=False, verbose=False, set_upstream=False, push_options=None, signed=None,
         force_with_lease=False, no_verify=False, repository=None, refspec=None,
         path=None):
    """ Push commits to a remote """

    cmd = ['push']

    if all:
        cmd.append('--all')

    if mirror:
        cmd.append('--mirror')

    if tags:
        cmd.append('--tags')

    if atomic:
        cmd.append('--atomic')

    if dry_run:
        cmd.append('--dry-run')

    if follow_tags:
        cmd.append('--follow-tags')

    if receive_pack:
        cmd.append(f'--receive-pack={receive_pack}')

    if repo:
        cmd.append(f'--repo={repo}')

    if force:
        cmd.append('--force')

    if delete:
        cmd.append('--delete')

    if prune:
        cmd.append('--prune')

    if verbose:
        cmd.append('--verbose')

    if set_upstream:
        cmd.append('--set-upstream')

    if push_options:
        for option in push_options:
            cmd.append(f'--push-option={option}')

    if signed:
        cmd.append(f'--signed={signed}')

    if force_with_lease:
        cmd.append('--force-with-lease')

    if no_verify:
        cmd.append('--no-verify')

    if repository:
        cmd.append(repository)

    if refspec:
        for ref in refspec:
            cmd.append(ref)

    return git(cmd, path=path)

################################################################################

def reset(sha1, path=None):
    """ Run git reset """

    return git(['reset', sha1], path=path)

################################################################################

def config_get(section, key, source=None, defaultvalue=None, path=None):
    """ Return the specified configuration entry
        Returns a default value if no matching configuration entry exists """

    cmd = ['config']

    if source == GLOBAL:
        cmd.append('--global')
    elif source == SYSTEM:
        cmd.append('--system')
    elif source == LOCAL:
        cmd.append('--local')
    elif source == WORKTREE:
        cmd.append('--worktree')

    cmd += ['--get', f'{section}.{key}']

    try:
        return git(cmd, path=path)[0]
    except GitError:
        return defaultvalue

################################################################################

def config_set(section, key, value, source=None, path=None):
    """ Set a configuration entry """

    cmd = ['config']

    if source == GLOBAL:
        cmd.append('--global')
    elif source == SYSTEM:
        cmd.append('--system')
    elif source == LOCAL:
        cmd.append('--local')
    elif source == WORKTREE:
        cmd.append('--worktree')

    cmd += ['--replace-all', f'{section}.{key}', value]

    return git(cmd, path=path)

################################################################################

def config_rm(section, key, source=LOCAL, path=None):
    """ Remove a configuration entry """

    cmd = ['config']

    if source == GLOBAL:
        cmd.append('--global')
    elif source == SYSTEM:
        cmd.append('--system')

    cmd += ['--unset', f'{section}.{key}']

    return git(cmd, path=path)

################################################################################

def ref(fields=('objectname',), sort=None, remotes=False, path=None):
    """ Wrapper for git for-each-ref """

    cmd = ['for-each-ref']

    if sort:
        cmd.append(f'--sort={sort}')

    field_list = []
    for field in fields:
        field_list.append('%%(%s)' % field)

    cmd += ['--format=%s' % '%00'.join(field_list), 'refs/heads']

    if remotes:
        cmd.append('refs/remotes/origin')

    for output in git(cmd, path=path):
        yield output.split('\0')

################################################################################

def branches(all=False, path=None, remote=False):
    """ Return a list of all the branches in the current repo """

    cmd = ['branch', '--format=%(refname:short)', '--list']

    if all:
        cmd.append('--all')

    if remote:
        cmd.append('--remote')

    results = []
    for output in git(cmd, path=path):
        if ' -> ' not in output and '(HEAD detached at ' not in output:
            results.append(output)

    return results

################################################################################

def delete_branch(branch, force=False, remote=False, path=None):
    """ Delete a branch, optionally forcefully and/or including the
        remote tracking branch """

    cmd = ['branch', '--delete']

    if force:
        cmd.append('--force')

    if remote:
        cmd.append('--remote')

    cmd.append(branch)

    return git(cmd, path=path)

################################################################################

def remote_prune(remote, dry_run=False, path=None):
    """ Return a list of remote tracking branches that no longer exist on the
        specified remote """

    cmd = ['remote', 'prune', remote]

    if dry_run:
        cmd.append('--dry-run')

    results = git(cmd, path=path)

    prunable_branches = []

    prune_re = re.compile(r'\s[*]\s\[would prune\]\s(.*)')

    for result in results:
        matches = prune_re.match(result)
        if matches:
            prunable_branches.append(matches.group(1))

    return prunable_branches

################################################################################

def get_commits(commit1, commit2, path=None):
    """ Get a list of commits separating two commits """

    return git(['rev-list', commit1, f'^{commit2}'], path=path)

################################################################################

def commit_count(commit1, commit2, path=None):
    """ Get a count of the number of commits between two commits """

    return int(git(['rev-list', '--count', commit1, f'^{commit2}'], path=path)[0])

################################################################################

def branch_name(branch, path=None):
    """ Return the full name of a branch given an abbreviation - e.g. @{upstream}
        for the upstream branch """

    return git(['rev-parse', '--abbrev-ref', '--symbolic-full-name', branch], path=path)[0]

################################################################################

def author(commit, path=None):
    """ Return the author of a commit """

    return git(['show', '--format=format:%an', commit], path=path)[0]

################################################################################

def commit_changes(commit='HEAD', path=None):
    """Return a list of the files changed in a commit"""

    return git(['show', '--name-only', '--pretty=format:', commit], path=path)

################################################################################

def files(dir=None, path=None):
    """ Return the output from 'git ls-files' """

    cmd = ['ls-files']

    if dir:
        cmd.append(dir)

    return git(cmd, path=path)

################################################################################

def stash(path=None):
    """ Return the list of stashed items (if any) """

    cmd = ['stash', 'list']

    return git(cmd, path=path)

################################################################################

def parents(commit=None, ignore=None, path=None):
    """ Look at the commits down the history of the specified branch,
        looking for another branch or branches that also contain the same commit.
        The first found is the parent (or equally-likely parents) of the
        branch - note due to fundamental git-ness a given branch can have multiple
        equally-plasuible parents.

        Return the list of possible parents and the distance down the branch
        from the current commit to those posible parents """

    # Get the history of the branch

    current_branch = commit or branch('HEAD', path=path)

    current_history = git(['rev-list', current_branch], path=path)

    # Look down the commits on the current branch for other branches that have
    # the same commit, using the ignore pattern if there is one.

    for distance, ancestor in enumerate(current_history):
        branches = []
        for brnch in git(['branch', '--contains', ancestor], path=path):
            brnch = brnch[2:]
            if brnch != current_branch and '(HEAD detached at' not in brnch:
                if not ignore or (ignore and not fnmatch.fnmatch(brnch, ignore)):
                    branches.append(brnch)

        if branches:
            break
    else:
        return None, 0

    return branches, distance

################################################################################

def find_common_ancestor(branch1='HEAD', branch2='master', path=None):
    """ Find the first (oldest) commit that the two branches have in common
        i.e. the point where one branch was forked from the other """

    return git(['merge-base', branch1, branch2], path=path)[0]

################################################################################

# List of boolean options to git.grep with corresponding command line option to use if option is True

_GREP_OPTLIST = \
    (
        ('color', '--color=always'),
        ('count', '--count'),
        ('follow', '--follow'),
        ('unmatch', '-I'),
        ('textconf', '--textconv'),
        ('ignore_case', '--ignore-case'),
        ('word_regexp', '--word-regexp'),
        ('invert_match', '--invert-match'),
        ('full_name', '--full-name'),
        ('extended_regexp', '--extended-regexp'),
        ('basic_regexp', '--basic-regexp'),
        ('perl_regexp', '--perl-regexp'),
        ('fixed_strings', '--fixed-strings'),
        ('line_number', '--line-number'),
        ('files_with_matches', '--files-with-matches'),
        ('files_without_match', '--files-without-match'),
        ('names_only', '--names-only'),
        ('null', '--null'),
        ('all_match', '--all-match'),
        ('quiet', '--quiet'),
        ('no_color', '--no-color'),
        ('break', '--break'),
        ('heading', '--heading'),
        ('show_function', '--show-function'),
        ('function_context', '--function-context'),
        ('only_matching', '--only-matching'),
    )

# List of non-boolean options to git.grep with corresponding command line option

_GREP_NON_BOOL_OPTLIST = \
    (
        ('root', '--root'),
        ('max_depth', '--max-depth'),
        ('after_context', '--after-context'),
        ('before_context', '--before-context'),
        ('context', '--context'),
        ('threads', '--threads'),
        ('file', '--file'),
        ('parent_basename', '--parent-basename')
    )

def grep(pattern, git_dir=None, work_tree=None, options=None, wildcards=None, path=None):
    """ Run git grep - takes a painfully large number of options passed
        as a dictionary. """

    cmd = []

    if git_dir:
        cmd += ['--git-dir', git_dir]

    if work_tree:
        cmd += ['--work-tree', work_tree]

    cmd += ['grep']

    if options:
        # Boolean options

        for opt in _GREP_OPTLIST:
            if options.get(opt[0], False):
                cmd.append(opt[1])

        # Non-boolean options

        for opt in _GREP_NON_BOOL_OPTLIST:
            value = options.get(opt[0], None)
            if value:
                cmd += (opt[1], value)

    if isinstance(pattern, list):
        cmd += pattern
    else:
        cmd += [pattern]

    if wildcards:
        cmd.append('--')
        cmd += wildcards

    return git_run_status(cmd, path=path)

################################################################################

def isbranch(branchname, path=None):
    """ Return true if the specified branch exists """

    return branchname in branches(True, path=path)

################################################################################

def default_branch(path=None):
    """ Return the name of the default branch, attempting to interrogate GitLab
        if the repo appears to have been cloned from there and falling back to
        returning whichever one of 'develop', 'main' or 'master' exists. """

    remote_list = remotes(path=path)
    if remote_list:
        for name in remote_list:
            if 'gitlab' in remote_list[name]:
                url = remote_list[name].split('@')[1].split(':')[0]
                repo = remote_list[name].split(':')[1]

                if not url.startswith('http://') and not url.startswith('https://'):
                    url = f'https://{url}'

                if repo.endswith('.git'):
                    repo = repo[:-4]

                try:
                    gl = gitlab.GitLab(url)
                    return gl.default_branch(repo)

                except gitlab.GitLabError:
                    return None

    git_branches = branches(path=path)
    default_branches = config_get('skilleter-thingy', 'defaultBranches', defaultvalue=DEFAULT_DEFAULT_BRANCHES, path=path).split(',')

    for branch in default_branches:
        if branch in git_branches:
            return branch

    raise GitError('Unable to determine default branch in the repo')

################################################################################

def matching_branch(branchname, case=False, path=None):
    """ Look for a branch matching the specified name and return it
        out if it is an exact match or there is only one partial
        match. If there are multiple branches that match, return them
        as a list.

        If case == True then the comparison is case-sensitive.

        If the branchname contains '*' or '?' wildcard matching is used,.
        otherwise, it just checks for a branches containing the branchname
        as a substring. """

    local_branches = branches(path=path)
    remote_branches = branches(path=path, remote=True)

    all_branches = local_branches + remote_branches

    # Always return exact matches

    if branchname in all_branches:
        return [branchname]

    matching = []
    matching_remote = []

    if not case:
        branchname = branchname.lower()

    if branchname == '-' * len(branchname):
        matching = [branchname]
    else:
        wildcard = '?' in branchname or '*' in branchname

        if wildcard:
            if branchname[0] not in ('?', '*'):
                branchname = f'*{branchname}'

            if branchname[-1] not in ('?', '*'):
                branchname = f'{branchname}*'

        for branch in all_branches:
            branchmatch = branch if case else branch.lower()

            # We have a partial match

            if (not wildcard and branchname in branchmatch) or (wildcard and fnmatch.fnmatch(branchmatch, branchname)):
                # If the match is a remote branch, ignore it if we already have the equivalent
                # local branch, otherwise add the name of the local branch that would be created.

                if branch in remote_branches:
                    localbranch = '/'.join(branch.split('/')[1:])
                    if localbranch not in matching:
                        matching_remote.append(localbranch)
                else:
                    matching.append(branch)

    # If we don't have matching local branches use the remote branch list (which may also be empty)

    if not matching:
        matching = matching_remote

    # Return the list of matches

    return matching

################################################################################

def update(clean=False, all=False, path=None):
    """ Run git update (which is a thingy command, and may end up as a module
        but for the moment, we'll treat it as any other git command) """

    cmd = ['update']

    if clean:
        cmd.append('--clean')

    if all:
        cmd.append('--all')

    return git(cmd, path=path)

################################################################################

def object_type(name, path=None):
    """ Return the git object type (commit, tag, blob, ...) """

    return git(['cat-file', '-t', name], path=path)[0]

################################################################################

def matching_commit(name, path=None):
    """ Similar to matching_branch() (see above).
        If the name uniquely matches a branch, return that
        If it matches multiple branches return a list
        If it doesn't match any branches, repeat the process for tags
        if it doesn't match any tags, repeat the process for commit IDs

        TODO: Currently matches multiple branches, tag, but only a unique commit - not sure if this the best behaviour, but it works """

    # First, look for exact matching object

    if iscommit(name, path=path):
        return [name]

    # Look for at least one matching branch

    matches = matching_branch(name, path=path)

    if matches:
        return matches

    # Look for at least one matching tag

    matches = []
    for tag in tags(path=path):
        if name in tag:
            matches.append(tag)

    if matches:
        return matches

    # Look for a matching commit

    try:
        commit_type = object_type(name, path=path)

        if commit_type == 'commit':
            matches = [name]
    except GitError:
        matches = []

    return matches

################################################################################

def log(branch1, branch2=None, path=None):
    """ Return the git log between the given commits """

    if branch2:
        cmd = ['log', f'{branch1}...{branch2}']
    else:
        cmd = ['log', '-n1', branch1]

    return git(cmd, path=path)

################################################################################

def clean(recurse=False, force=False, dry_run=False, quiet=False,
          exclude=None, ignore_rules=False, remove_only_ignored=False, path=None):
    """ Run git clean """

    cmd = ['clean']

    if recurse:
        cmd.append('-d')

    if force:
        cmd.append('--force')

    if dry_run:
        cmd.append('--dry-run')

    if quiet:
        cmd.append('--quiet')

    if exclude:
        cmd += ['--exclude', exclude]

    if ignore_rules:
        cmd.append('-x')

    if remove_only_ignored:
        cmd.append('-X')

    return git(cmd, path=path)

################################################################################

def run_tests():
    """Test suite for the module"""

    print('Creating local git repo')

    init('test-repo')
    os.chdir('test-repo')

    try:
        print('Initial status:')
        print('  User name:            %s' % config_get('user', 'name'))
        print('  Project:              %s' % project())
        print('  Remotes:              %s' % remotes())
        print('  Current branch:       %s' % branch())
        print('  Current working tree: %s' % working_tree())
        print('  Rebasing?             %s' % rebasing())
        print('  Status:               %s' % status())
        print()

        print('Adding a removing a configuration value')

        config_set('user', 'size', 'medium')

        print('  User size config:     %s' % config_get('user', 'size'))

        config_rm('user', 'size')

        value = config_get('user', 'size')
        if value is None:
            print('  Successfully failed to read the newly-deleted config data')
        else:
            raise GitError('Unexpected lack of error reading configuration data', 1)

        config_set('user', 'email', 'user@localhost')
        config_set('user', 'name', 'User Name')

        print('')

        with open('newfile.txt', 'w', encoding='utf8') as newfile:
            newfile.write('THIS IS A TEST')

        print('Adding and committing "newfile.txt"')

        add(['newfile.txt'])

        commit(['newfile.txt'], 'Test the add and commit functionality')

        print('Removing and committing "newfile.txt"')

        rm(['newfile.txt'])

        commit(None, 'Removed "newfile.txt"')

        print('Removing the last commit')

        reset('HEAD~1')

        print('Commit info for HEAD  %s' % commit_info('HEAD'))

    except GitError as exc:
        sys.stderr.write(f'ERROR: {exc.msg}')
        sys.exit(1)

    finally:
        # If anything fails, then clean up afterwards

        os.chdir('..')
        shutil.rmtree('test-repo')

################################################################################
# Entry point

if __name__ == '__main__':
    run_tests()



import os

import skilleter_modules.files as files


def test_format_size_suffix_and_sign():
    assert files.format_size(-1024, always_suffix=True).startswith('-')
    assert files.format_size(1024, always_suffix=True).endswith(' KiB')


def test_backup_no_file_no_raise(tmp_path):
    missing = tmp_path / 'missing.txt'
    # Should no-op when file is absent
    files.backup(str(missing))
    assert not list(tmp_path.iterdir())


def test_backup_creates_copy(tmp_path):
    src = tmp_path / 'file.txt'
    src.write_text('data', encoding='utf-8')
    files.backup(str(src))
    assert (tmp_path / 'file.bak').read_text(encoding='utf-8') == 'data'



import skilleter_modules.path as path


def test_is_subdirectory_strict_descendant():
    assert path.is_subdirectory('/root', '/root/child') is True


def test_is_subdirectory_same_path_is_false():
    assert path.is_subdirectory('/root', '/root') is False


def test_is_subdirectory_unrelated_is_false():
    assert path.is_subdirectory('/root/one', '/root/two') is False


def test_trimpath_home_to_tilde(monkeypatch):
    monkeypatch.setenv('HOME', '/home/user')
    assert path.trimpath('/home/user/projects', 80).startswith('~/')


def test_trimpath_truncates_middle():
    trimmed = path.trimpath('/a/b/c/d/e/f/g/h/i', 10)
    assert '...' in trimmed



""" Default dircolors data, and the reference .dircolors file.
Generated from dircolors in GNU coreutils version 8.31 """

__all__ = ['DEFAULT_DIRCOLORS', 'DEFAULT_LS_COLORS']

# the output of `dircolors -p`
DEFAULT_DIRCOLORS = r"""
# Configuration file for dircolors, a utility to help you set the
# LS_COLORS environment variable used by GNU ls with the --color option.
# Copyright (C) 1996-2019 Free Software Foundation, Inc.
# Copying and distribution of this file, with or without modification,
# are permitted provided the copyright notice and this notice are preserved.
# The keywords COLOR, OPTIONS, and EIGHTBIT (honored by the
# slackware version of dircolors) are recognized but ignored.
# Below are TERM entries, which can be a glob patterns, to match
# against the TERM environment variable to determine if it is colorizable.
TERM Eterm
TERM ansi
TERM *color*
TERM con[0-9]*x[0-9]*
TERM cons25
TERM console
TERM cygwin
TERM dtterm
TERM gnome
TERM hurd
TERM jfbterm
TERM konsole
TERM kterm
TERM linux
TERM linux-c
TERM mlterm
TERM putty
TERM rxvt*
TERM screen*
TERM st
TERM terminator
TERM tmux*
TERM vt100
TERM xterm*
# Below are the color init strings for the basic file types.
# One can use codes for 256 or more colors supported by modern terminals.
# The default color codes use the capabilities of an 8 color terminal
# with some additional attributes as per the following codes:
# Attribute codes:
# 00=none 01=bold 04=underscore 05=blink 07=reverse 08=concealed
# Text color codes:
# 30=black 31=red 32=green 33=yellow 34=blue 35=magenta 36=cyan 37=white
# Background color codes:
# 40=black 41=red 42=green 43=yellow 44=blue 45=magenta 46=cyan 47=white
#NORMAL 00 # no color code at all
#FILE 00 # regular file: use no color at all
RESET 0 # reset to "normal" color
DIR 01;34 # directory
LINK 01;36 # symbolic link. (If you set this to 'target' instead of a
 # numerical value, the color is as for the file pointed to.)
MULTIHARDLINK 00 # regular file with more than one link
FIFO 40;33 # pipe
SOCK 01;35 # socket
DOOR 01;35 # door
BLK 40;33;01 # block device driver
CHR 40;33;01 # character device driver
ORPHAN 40;31;01 # symlink to nonexistent file, or non-stat'able file ...
MISSING 00 # ... and the files they point to
SETUID 37;41 # file that is setuid (u+s)
SETGID 30;43 # file that is setgid (g+s)
CAPABILITY 30;41 # file with capability
STICKY_OTHER_WRITABLE 30;42 # dir that is sticky and other-writable (+t,o+w)
OTHER_WRITABLE 34;42 # dir that is other-writable (o+w) and not sticky
STICKY 37;44 # dir with the sticky bit set (+t) and not other-writable
# This is for files with execute permission:
EXEC 01;32
# List any file extensions like '.gz' or '.tar' that you would like ls
# to colorize below. Put the extension, a space, and the color init string.
# (and any comments you want to add after a '#')
# If you use DOS-style suffixes, you may want to uncomment the following:
#.cmd 01;32 # executables (bright green)
#.exe 01;32
#.com 01;32
#.btm 01;32
#.bat 01;32
# Or if you want to colorize scripts even if they do not have the
# executable bit actually set.
#.sh 01;32
#.csh 01;32
 # archives or compressed (bright red)
.tar 01;31
.tgz 01;31
.arc 01;31
.arj 01;31
.taz 01;31
.lha 01;31
.lz4 01;31
.lzh 01;31
.lzma 01;31
.tlz 01;31
.txz 01;31
.tzo 01;31
.t7z 01;31
.zip 01;31
.z 01;31
.dz 01;31
.gz 01;31
.lrz 01;31
.lz 01;31
.lzo 01;31
.xz 01;31
.zst 01;31
.tzst 01;31
.bz2 01;31
.bz 01;31
.tbz 01;31
.tbz2 01;31
.tz 01;31
.deb 01;31
.rpm 01;31
.jar 01;31
.war 01;31
.ear 01;31
.sar 01;31
.rar 01;31
.alz 01;31
.ace 01;31
.zoo 01;31
.cpio 01;31
.7z 01;31
.rz 01;31
.cab 01;31
.wim 01;31
.swm 01;31
.dwm 01;31
.esd 01;31
# image formats
.jpg 01;35
.jpeg 01;35
.mjpg 01;35
.mjpeg 01;35
.gif 01;35
.bmp 01;35
.pbm 01;35
.pgm 01;35
.ppm 01;35
.tga 01;35
.xbm 01;35
.xpm 01;35
.tif 01;35
.tiff 01;35
.png 01;35
.svg 01;35
.svgz 01;35
.mng 01;35
.pcx 01;35
.mov 01;35
.mpg 01;35
.mpeg 01;35
.m2v 01;35
.mkv 01;35
.webm 01;35
.ogm 01;35
.mp4 01;35
.m4v 01;35
.mp4v 01;35
.vob 01;35
.qt 01;35
.nuv 01;35
.wmv 01;35
.asf 01;35
.rm 01;35
.rmvb 01;35
.flc 01;35
.avi 01;35
.fli 01;35
.flv 01;35
.gl 01;35
.dl 01;35
.xcf 01;35
.xwd 01;35
.yuv 01;35
.cgm 01;35
.emf 01;35
# https://wiki.xiph.org/MIME_Types_and_File_Extensions
.ogv 01;35
.ogx 01;35
# audio formats
.aac 00;36
.au 00;36
.flac 00;36
.m4a 00;36
.mid 00;36
.midi 00;36
.mka 00;36
.mp3 00;36
.mpc 00;36
.ogg 00;36
.ra 00;36
.wav 00;36
# https://wiki.xiph.org/MIME_Types_and_File_Extensions
.oga 00;36
.opus 00;36
.spx 00;36
.xspf 00;36
"""

# the above converted to LS_COLORS by dircolors, but with the trailing ':' stripped
DEFAULT_LS_COLORS = r"""
rs=0:di=01;34:ln=01;36:mh=00:pi=40;33:so=01;35:do=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:
mi=00:su=37;41:sg=30;43:ca=30;41:tw=30;42:ow=34;42:st=37;44:ex=01;32:*.tar=01;31:*.tgz=01;31:
*.arc=01;31:*.arj=01;31:*.taz=01;31:*.lha=01;31:*.lz4=01;31:*.lzh=01;31:*.lzma=01;31:*.tlz=01;31:
*.txz=01;31:*.tzo=01;31:*.t7z=01;31:*.zip=01;31:*.z=01;31:*.dz=01;31:*.gz=01;31:*.lrz=01;31:
*.lz=01;31:*.lzo=01;31:*.xz=01;31:*.zst=01;31:*.tzst=01;31:*.bz2=01;31:*.bz=01;31:*.tbz=01;31:
*.tbz2=01;31:*.tz=01;31:*.deb=01;31:*.rpm=01;31:*.jar=01;31:*.war=01;31:*.ear=01;31:*.sar=01;31:
*.rar=01;31:*.alz=01;31:*.ace=01;31:*.zoo=01;31:*.cpio=01;31:*.7z=01;31:*.rz=01;31:*.cab=01;31:
*.wim=01;31:*.swm=01;31:*.dwm=01;31:*.esd=01;31:*.jpg=01;35:*.jpeg=01;35:*.mjpg=01;35:
*.mjpeg=01;35:*.gif=01;35:*.bmp=01;35:*.pbm=01;35:*.pgm=01;35:*.ppm=01;35:*.tga=01;35:*.xbm=01;35:
*.xpm=01;35:*.tif=01;35:*.tiff=01;35:*.png=01;35:*.svg=01;35:*.svgz=01;35:*.mng=01;35:*.pcx=01;35:
*.mov=01;35:*.mpg=01;35:*.mpeg=01;35:*.m2v=01;35:*.mkv=01;35:*.webm=01;35:*.ogm=01;35:*.mp4=01;35:
*.m4v=01;35:*.mp4v=01;35:*.vob=01;35:*.qt=01;35:*.nuv=01;35:*.wmv=01;35:*.asf=01;35:*.rm=01;35:
*.rmvb=01;35:*.flc=01;35:*.avi=01;35:*.fli=01;35:*.flv=01;35:*.gl=01;35:*.dl=01;35:*.xcf=01;35:
*.xwd=01;35:*.yuv=01;35:*.cgm=01;35:*.emf=01;35:*.ogv=01;35:*.ogx=01;35:*.aac=00;36:*.au=00;36:
*.flac=00;36:*.m4a=00;36:*.mid=00;36:*.midi=00;36:*.mka=00;36:*.mp3=00;36:*.mpc=00;36:*.ogg=00;36:
*.ra=00;36:*.wav=00;36:*.oga=00;36:*.opus=00;36:*.spx=00;36:*.xspf=00;36""".replace('\n', '')



"""Minimal test code for colour"""

import io

import skilleter_modules.colour as colour

def test_colour():
    """Very basic test"""

    for combo in (0, 1, 2):
        print()
        if combo == 0:
            print('Background colours')
        elif combo == 1:
            print('Foreground colours')
        else:
            print('Combinations')

        print()
        for y in range(0, 16):
            for x in range(0, 16):
                colour_index = x + y * 16

                if combo == 0:
                    colour.write(format('[B%d]%4d' % (colour_index, colour_index)), newline=False)
                elif combo == 1:
                    colour.write(format('[%d]%4d' % (colour_index, colour_index)), newline=False)
                else:
                    colour.write(format('[B%d]%4d[%d]/%4d ' % (colour_index, colour_index, 255 - colour_index, 255 - colour_index)), newline=False)

            colour.write('[NORMAL]')

    print()

    colour.write('Foreground: [RED]red [GREEN]green [BLACK]black [NORMAL]normal')
    colour.write('Background: [BRED]red [BGREEN]green [BBLACK]black [NORMAL]normal')

    colour.write('Foreground: [BBLUE:blue] [RED:red] normal')

    colour.write('Bright foreground: [RED_B]red [GREEN_B]green [BLACK_B]black [NORMAL]normal')
    colour.write('Bright background: [BRED_B]red [BGREEN_B]green [BBLACK_B]black [NORMAL]normal')

    colour.write('Foreground: [BBLUE:blue_B] [RED:red_B] normal')

    print()

    colour.write('[NORMAL:Normal text]')
    colour.write('[FAINT:Faint text]')
    colour.write('[ITALIC:Italic text]')
    colour.write('[UNDERSCORE:Underscored text]')
    colour.write('[BLINK:Blinking text]')
    colour.write('[REVERSE:Reverse text]')
    colour.write('[STRIKE:Strikethrough text]')


def test_write_empty_string_no_indent():
    buf = io.StringIO()
    colour.write('', indent=4, stream=buf)
    assert buf.getvalue() == '\n'


def test_write_with_indent_and_strip():
    buf = io.StringIO()
    colour.write('  text', indent=2, strip=True, stream=buf)
    assert buf.getvalue() == '  text\n'



#! /usr/bin/env python3

################################################################################
""" Thingy file and directory functionality

    Copyright (C) 2017-18 John Skilleter

    Licence: GPL v3 or later
"""
################################################################################

import os
import logging

################################################################################

class PathError(Exception):
    """ Exception raised by the module """

    def __init__(self, msg):
        super(PathError, self).__init__(msg)
        self.msg = msg

################################################################################

def is_subdirectory(root_path, sub_path):
    """ Return True if sub_path is a sub-directory of root_path """

    abs_sub_path = os.path.abspath(sub_path)
    abs_root_path = os.path.abspath(root_path)

    logging.debug('root path: %s', abs_root_path)
    logging.debug('sub path : %s', abs_sub_path)

    common = os.path.commonpath([abs_root_path, abs_sub_path])

    # Require a strict subdirectory: common path matches root and paths differ
    return common == abs_root_path and abs_sub_path != abs_root_path

################################################################################

def trimpath(full_path, trim_width):
    """ Trim a path to a specified maximum width, but always leaving the
        lowest-level directory (even if it exceeds the trim width). """

    logging.debug('Path: "%s"', full_path)
    logging.debug('Required width: %d', trim_width)

    full_path = os.path.abspath(full_path)

    # Remove any trailing '/' from the path

    if full_path != '/' and full_path[-1] == '/':
        full_path = full_path[:-1]

    # If the path starts with the user's home directory then convert the prefix
    # into a '~'

    home_dir = os.path.expanduser('~')

    if full_path == home_dir:
        full_path = '~'
        logging.debug('Converted path to "~"')

    elif is_subdirectory(home_dir, full_path):
        full_path = "~/%s" % full_path[len(home_dir) + 1:]

        logging.debug('Converted path to "%s"', full_path)

    # If the path is too long then slice it into directories and cut sub-directories
    # out of the middle until it is short enough. Always leave the last element
    # in place, even if this means total length exceeds the requirement.

    path_len = len(full_path)

    logging.debug('Path length: %d', path_len)

    # Already within maximum width, so just return it

    if path_len <= trim_width:
        return full_path

    # Split into an array of directories and trim out middle ones

    directories = full_path.split('/')

    logging.debug('Path has %d elements: "%s"', len(directories), directories)

    if len(directories) == 1:
        # If there's only one element in the path, just give up

        logging.debug('Only 1 directory in the path, leaving it as-is')

    elif len(directories) == 2:
        # If there's only two elements in the path then replace the first
        # element with '...' and give up

        logging.debug('Only 2 directories in the path, so setting the first to "..."')

        directories[0] = '...'

    else:
        # Start in the middle and remove entries to the left and right until the total
        # path length is shortened to a sufficient extent

        right = len(directories) // 2
        left = right - 1
        first = True

        while path_len > trim_width:

            path_len -= len(directories[right]) + 1

            if first:
                path_len += 4
                first = False

            if right == len(directories) - 1:
                break

            right += 1

            if path_len > trim_width:
                path_len -= len(directories[left]) + 1

                if left == 0:
                    break

                left -= 1

        logging.debug('Removing entries %d..%d from the path', left, right)

        directories = directories[0:left + 1] + ['...'] + directories[right:]

    full_path = '/'.join(directories)

    logging.debug('Calculated width is %d and actual width is %d', path_len, len(full_path))

    return full_path

################################################################################

if __name__ == '__main__':
    PARENT = '/1/2/3/5'
    CHILD = '/1/2/3/5/6'

    print('Is %s a subdirectory of %s: %s (expecting True)' % (CHILD, PARENT, is_subdirectory(PARENT, CHILD)))
    print('Is %s a subdirectory of %s: %s (expecting False)' % (PARENT, CHILD, is_subdirectory(CHILD, PARENT)))

    LONG_PATH = '/home/jms/source/womble-biscuit-token-generation-service/subdirectory'

    for pathname in (LONG_PATH, os.path.realpath('.')):
        print('Full path: %s' % pathname)

        for length in (80, 60, 40, 20, 16, 10):
            print('Trimmed to %d characters: %s' % (length, trimpath(pathname, length)))



import skilleter_modules.tidy as tidy


def test_remove_times_does_not_match_stray_dot():
    text = 'Version 1. release'
    assert tidy.remove_times(text) == text


def test_remove_times_replaces_seconds():
    assert tidy.remove_times('took 12.5 seconds') == 'took {ELAPSED}'



import types

import pytest

import skilleter_modules.docker as docker


class DummyResult(types.SimpleNamespace):
    pass


def test_instances_parses_lines(monkeypatch):
    def fake_run(cmd, capture_output, check, text):
        assert cmd[:2] == ['docker', 'ps']
        return DummyResult(stdout='id1\nid2\n')

    monkeypatch.setattr('subprocess.run', fake_run)
    assert docker.instances() == ['id1', 'id2']


def test_stop_force_passes_flag(monkeypatch):
    seen = {}

    def fake_run(cmd, check, capture_output):
        seen['cmd'] = cmd
        return DummyResult()

    monkeypatch.setattr('subprocess.run', fake_run)
    docker.stop('abc', force=True)
    assert seen['cmd'] == ['docker', 'stop', '--force', 'abc']



import io
import stat

import skilleter_modules.dircolors as dircolors


def test_load_from_dircolors_does_not_close_stream():
    data = io.StringIO('DIR 01;34\n')
    dc = dircolors.Dircolors(load=False)
    assert dc.load_from_dircolors(data, strict=True) is True
    assert data.closed is False
    assert dc.loaded is True


def test_format_mode_directory_and_extension(tmp_path):
    dc = dircolors.Dircolors(load=False)
    # Load minimal rules: directories blue (01;34), *.txt red (31)
    rules = 'DIR 01;34\n*.txt 31\n'
    assert dc.load_from_dircolors(io.StringIO(rules), strict=True)

    # Directory formatting: should wrap with escape codes
    formatted_dir = dc.format_mode('name', (stat.S_IFDIR | stat.S_IRUSR))
    assert '\x1b[' in formatted_dir and formatted_dir.endswith('\x1b[0m')

    # Extension formatting: .txt uses red foreground
    formatted_file = dc.format_mode('file.txt', 0)
    assert '\x1b[' in formatted_file and formatted_file.endswith('\x1b[0m')



import subprocess

import skilleter_modules.run as run


def test_process_echoes_and_captures_stdout(monkeypatch):
    real_popen = subprocess.Popen

    def fake_popen(command, bufsize=0, stdout=None, stderr=None, text=False, errors=None, encoding=None, **kwargs):
        assert stdout == subprocess.PIPE
        assert stderr == subprocess.PIPE
        proc = real_popen(['printf', 'hello\n'], stdout=stdout, stderr=stderr, text=True)
        return proc

    monkeypatch.setattr(subprocess, 'Popen', fake_popen)
    result = run._process(['printf', 'hello'])
    assert result['stdout'] == ['hello']


def test_process_raises_on_nonzero(monkeypatch):
    real_popen = subprocess.Popen

    def fake_popen(command, bufsize=0, stdout=None, stderr=None, text=False, errors=None, encoding=None, **kwargs):
        return real_popen(['python', '-c', 'import sys; sys.exit(3)'], stdout=stdout, stderr=stderr, text=True)

    monkeypatch.setattr(subprocess, 'Popen', fake_popen)
    try:
        run._process(['python', '-c', 'exit 3'])
    except run.RunError as exc:
        assert exc.status == 3
    else:
        assert False, 'expected RunError'



#! /usr/bin/env python3

""" Convert colour highlighting codes from the LS_COLORS environment variable
    used by ls to curses
"""

################################################################################

import sys
import os
import glob
import fnmatch
import curses
import stat

################################################################################

class CursesDircolors:
    """ Convert dircolors codes to curses colours """

    # Convert standard foreground and background codes to curses equivalents

    ANSI_CONVERT_FORE = {
        30: curses.COLOR_BLACK,
        31: curses.COLOR_RED,
        32: curses.COLOR_GREEN,
        33: curses.COLOR_YELLOW,
        34: curses.COLOR_BLUE,
        35: curses.COLOR_MAGENTA,
        36: curses.COLOR_CYAN,
        37: curses.COLOR_WHITE,
    }

    ANSI_CONVERT_BACK = {
        40: curses.COLOR_BLACK,
        41: curses.COLOR_RED,
        42: curses.COLOR_GREEN,
        43: curses.COLOR_YELLOW,
        44: curses.COLOR_BLUE,
        45: curses.COLOR_MAGENTA,
        46: curses.COLOR_CYAN,
        47: curses.COLOR_WHITE,
    }

    # Convert attribute codes to their meanings
    # TODO: Attributes not handled yet

    ANSI_CONVERT_ATTR = {
        0: 0,
        1: curses.A_BOLD,
        4: curses.A_UNDERLINE,
        5: curses.A_BLINK,
        7: curses.A_BLINK,
        8: curses.A_INVIS,
    }

    # Default colour

    DEFAULT_ATTR = {'attr': [], 'fore': -1, 'back': -1}

    ################################################################################

    def __init__(self, reserved=0):
        # Create the lookup tables associating special type codes or wildcards
        # with colour pairs.

        self.colour_pairs = [[-1, -1]]

        self.wildcard_highlight = {}
        self.special_highlight = {}

        self.reserved = reserved

        self.init_ls_colours()

    ################################################################################

    def curses_alloc_pair(self, attr):
        """ Given a set of attributes return the equivalent curses colour pair,
            creating a new one if a matching one doesn't already exsit """

        colours = [attr['fore'], attr['back']]

        # Get an existing colour pair that uses the same colours or create
        # a new one if one doesn't exist. If a pair slot is already allocated
        # but the curses subsystem cannot allocate (e.g., no colours), skip init.

        if colours in self.colour_pairs:
            pair_index = self.colour_pairs.index(colours) + self.reserved
        else:
            pair_index = len(self.colour_pairs) + self.reserved
            self.colour_pairs.append(colours)
            try:
                curses.init_pair(pair_index, attr['fore'], attr['back'])
            except curses.error:
                pass

        return pair_index

    ################################################################################

    def curses_colour(self, code):
        """ Return a cursors colour pair index for the specified dircolor colour
            code string. """

        # Default attribute

        attr = {'attr': [], 'fore': -1, 'back': -1}

        # Non-zero if processing multi-value colour code

        special = 0
        special_item = None

        # We trigger a ValueError and fail on anything that's wrong in the code

        try:
            # Split into fields and convert to integer values

            codes = [int(c) for c in code.split(';')]

            for entry in codes:
                # Process 2nd entry in a special colour sequence - must have value of 5

                if special == 1:
                    if entry != 5:
                        raise ValueError
                    special = 2

                # Process 3rd entry in a special colour sequence - must be the colour
                # code between 0 and 255

                elif special == 2:
                    if entry < 0 or entry > 255:
                        raise ValueError

                    attr[special_item] = entry
                    special = 0

                # Normal foreground colour

                elif entry in self.ANSI_CONVERT_FORE:
                    attr['fore'] = self.ANSI_CONVERT_FORE[entry]

                # Normal background colour

                elif entry in self.ANSI_CONVERT_BACK:
                    attr['back'] = self.ANSI_CONVERT_BACK[entry]

                # Special foreground colour in the form 38;5;VALUE

                elif entry == 38:
                    special = 1
                    special_item = 'fore'

                # Special background colour in the form 48;5;VALUE

                elif entry == 48:
                    special = 1
                    special_item = 'back'

                # Attribute (underline, bold, etc.)

                elif entry in self.ANSI_CONVERT_ATTR:
                    attr['attr'].append(self.ANSI_CONVERT_ATTR[entry])

                # Anything else is an error

                else:
                    raise ValueError

        except ValueError:
            print(f'Invalid colour specification: "{code}"')
            sys.exit(1)

        # Allocate a colour pair for the colour combination and return it

        return self.curses_alloc_pair(attr)

    ################################################################################

    def init_ls_colours(self):
        """ Generate tables matching special file types (fifos, sockets, etc.) and
            wildcards to curses colour pairs """

        colour_data = os.environ.get('LS_COLORS', '').split(':')

        # Iterate through the highlighters, create/get a colour pair corresponding
        # to the colour codes and save one of the tables.

        for item in colour_data:
            item = item.strip()
            if '=' in item:
                code, colour = item.split('=')

                colour_pair = self.curses_colour(colour)

                if len(code) == 2 and '*' not in code and '.' not in code:
                    self.special_highlight[code] = colour_pair
                else:
                    self.wildcard_highlight[code] = colour_pair

    ################################################################################

    def get_colour(self, filename, filemode=None):
        """ Get the curses colour for a filename, returns 0 if no highlighting
            is needed """

        if filemode:
            if stat.S_ISDIR(filemode):
                if 'di' in self.special_highlight:
                    return self.special_highlight['di']
            elif stat.S_ISLNK(filemode):
                destfile = os.readlink(filename)

                if os.path.exists(destfile):
                    if 'ln' in self.special_highlight:
                        return self.special_highlight['ln']
                elif 'or' in self.special_highlight:
                    return self.special_highlight['or']

            elif stat.S_ISBLK(filemode):
                if 'bd' in self.special_highlight:
                    return self.special_highlight['bd']
            elif stat.S_ISCHR(filemode):
                if 'cd' in self.special_highlight:
                    return self.special_highlight['cd']

            if filemode & stat.S_IXUSR:
                if 'ex' in self.special_highlight:
                    return self.special_highlight['ex']

        for entry in self.wildcard_highlight:
            if fnmatch.fnmatch(filename, entry):
                colour = self.wildcard_highlight[entry]
                break
        else:
            colour = 0

        return colour

    ################################################################################

    def get_colour_pair(self, filename, filemode=None):
        """ Get the curses colour pair for a filename, optionally specifying the
            file mode (as per os.stat()) """

        return curses.color_pair(self.get_colour(filename, filemode))

################################################################################

def _test_code(stdscr):
    """ Entry point """

    curses.start_color()
    curses.use_default_colors()

    # Initialise colours

    dc = CursesDircolors()

    # Demo code to list files specified by the first command line argument
    # highlighted appropriately

    y = 0
    for filename in glob.glob(sys.argv[1]):
        colour = dc.get_colour(filename)

        stdscr.addstr(y, 0, filename, curses.color_pair(colour))

        y += 1
        if y > 30:
            break

    stdscr.getch()

################################################################################

if __name__ == "__main__":
    curses.wrapper(_test_code)



################################################################################
""" Curses-based pop-up message

Usage:
    with PopUp(curses_screen, message, colour):
        do_stuff

Popup message is displayed for the duration of the with statement, and
has optional parameters to wait for a keypress, and/or pause before removing
the popup again.

"""
################################################################################

import time
import curses
import curses.panel

################################################################################

class PopUp():
    """ Class to enable popup windows to be used via with statements """

    def __init__(self, screen, msg, colour, waitkey=False, sleep=True, centre=True, refresh=True):
        """ Initialisation - just save the popup parameters """

        self.panel = None
        self.screen = screen
        self.msg = msg
        self.centre = centre
        self.colour = curses.color_pair(colour)
        self.refresh = refresh
        self.sleep = sleep and not waitkey
        self.waitkey = waitkey
        self.start_time = 0

    def __enter__(self):
        """ Display the popup """

        lines = self.msg.split('\n')
        height = len(lines)

        width = 0
        for line in lines:
            width = max(width, len(line))

        width += 2
        height += 2

        size_y, size_x = self.screen.getmaxyx()

        window = curses.newwin(height, width, (size_y - height) // 2, (size_x - width) // 2)
        self.panel = curses.panel.new_panel(window)

        window.bkgd(' ', self.colour)
        for y_pos, line in enumerate(lines):
            x_pos = (width - len(line)) // 2 if self.centre else 1
            window.addstr(y_pos + 1, x_pos, line, self.colour)

        self.panel.top()
        curses.panel.update_panels()
        self.screen.refresh()

        self.start_time = time.monotonic()

        return self

    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
        """ Remove the popup """

        if self.waitkey:
            while True:
                keypress = self.screen.getch()
                if keypress != curses.KEY_RESIZE:
                    break

                curses.panel.update_panels()
                self.screen.refresh()

        if self.panel:
            if self.sleep:
                elapsed = time.monotonic() - self.start_time

                if elapsed < 1:
                    time.sleep(1 - elapsed)

            del self.panel

        if self.refresh:
            self.screen.refresh()



#! /usr/bin/env python3

################################################################################
""" Code for running a subprocess, optionally capturing stderr and/or stdout and
    optionally echoing either or both to the console in realtime and storing.

    Uses threads to capture and output stderr and stdout since this seems to be
    the only way to do it (Popen does not have the ability to output the process
    stdout output to the stdout output).

    Intended for more versatile replacement for the thingy process.run() function
    which can handle all combinations of foreground/background console/return
    stderr/stdout/both options. """

# TODO: This does not run on Python versions <3.5 (so Ubuntu 14.04 is a problem!)
################################################################################

################################################################################
# Imports

import sys
import subprocess
import threading
import shlex

from . import tidy

################################################################################

class RunError(Exception):
    """ Run exception """

    def __init__(self, msg, status=1):
        super().__init__(msg)
        self.msg = msg
        self.status = status

################################################################################

# TODO: This is the _process() and run() replacement once additional parameters have been implemented
# as those functions are probably over-specified, so need to work out what functionality is ACTUALLY being used!

def command(cmd, show_stdout=False, show_stderr=False):
    """
    Run an external command and optionally stream its stdout/stderr to the console
    while capturing them.

    Args:
        cmd: Command to run (string or argv list). If a string, it will be split with shlex.split().
        show_stdout: If True, echo stdout lines to the console as they arrive.
        show_stderr: If True, echo stderr lines to the console as they arrive.

    Returns:
        (returncode, stdout_lines, stderr_lines)
    """

    def _pump(stream, sink, echo):
        """Thread to capture and optionally echo output from subprocess.Popen"""

        # Read line-by-line until EOF

        for line in iter(stream.readline, ""):
            # Strip trailing newline when storing; keep original when echoing
            sink.append(line.rstrip("\n"))

            if echo:
                print(tidy.convert_ansi(line), end="", flush=True)

        stream.close()

    # Normalize command to be a string

    if isinstance(cmd, str):
        cmd = shlex.split(cmd)

    # Storage for stdout/stderr

    stdout_lines = []
    stderr_lines = []

    # Start process with separate pipes; line-buffering for timely reads

    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,                 # decode to str
        bufsize=1,                 # line-buffered
        universal_newlines=True,   # compatibility alias
        errors="replace"           # avoid crashes on decoding issues
    )

    # Threads to read both streams concurrently (prevents deadlocks)

    t_out = threading.Thread(target=_pump, args=(proc.stdout, stdout_lines, show_stdout))
    t_err = threading.Thread(target=_pump, args=(proc.stderr, stderr_lines, show_stderr))

    t_out.start()
    t_err.start()

    # Wait for process to complete and threads to drain

    returncode = proc.wait()
    t_out.join()
    t_err.join()

    # Return the status, stdout and stderr

    return returncode, stdout_lines, stderr_lines

################################################################################

def capture_output(cmd, input_stream, output_streams):
    """ Capture data from a stream (input_stream), optionally
        outputting it (if output_streams is not None and optionally
        saving it into a variable (data, if not None), terminating
        when the specified command (cmd, which is presumed to be the process
        outputting to the input stream) exits.
        TODO: Use of convert_ansi should be controlled via a parameter (off/light/dark)
        TODO: Another issue is that readline() only returns at EOF or EOL, so if you get a prompt "Continue?" with no newline you do not see it until after you respond to it.
    """

    while True:
        output = input_stream.readline()

        if output:
            if output_streams:
                for stream in output_streams:
                    if isinstance(stream, list):
                        stream.append(output.rstrip())
                    else:
                        if stream in (sys.stdout, sys.stderr):
                            stream.write(tidy.convert_ansi(output))
                        else:
                            stream.write(output)

        elif cmd.poll() is not None:
            return

################################################################################

def _process(command,
             stdout=None, stderr=None,
             output=None):
    """ Run an external command.

        stdout and stderr indicate whether stdout/err are output and/or sent to a file and/or stored in a variable.
        They can be boolean (True: output to sys.stdout/err, False: Do nothing), a file handle or a variable, or an
        array of any number of these (except booleans).

        If output is True then stdout and stderr are both output as if stdout=True and stderr=True (in addition to
        any other values passed in those parameters)

        The return value is a tuple consisting of the status code, captured stdout (if any) and captured
        stderr (if any).

        Will raise OSError if the command could not be run and RunError if the command returned a non-zero status code.
    """

    # If stdout/stderr are booleans then output to stdout/stderr if True, else discard output

    if isinstance(stdout, bool):
        stdout = sys.stdout if stdout else None

    if isinstance(stderr, bool):
        stderr = sys.stderr if stderr else None

    # If stdout/stderr are not arrays then make them so

    if not isinstance(stdout, list):
        stdout = [stdout] if stdout else []

    if not isinstance(stderr, list):
        stderr = [stderr] if stderr else []

    # If output is True then add stderr/out to the list of outputs

    if output:
        if sys.stdout not in stdout:
            stdout.append(sys.stdout)

        if sys.stderr not in stderr:
            stderr.append(sys.stderr)

    # Capture stdout/stderr to arrays

    stdout_data = []
    stderr_data = []

    stdout.append(stdout_data)
    stderr.append(stderr_data)

    if isinstance(command, str):
        command = shlex.split(command, comments=True)

    # Use a pipe for stdout/stderr if are are capturing it
    # and send it to /dev/null if we don't care about it at all.

    if stdout == [sys.stdout] and not stderr:
        stdout_stream = subprocess.STDOUT
        stderr_stream = subprocess.DEVNULL
    else:
        stdout_stream = subprocess.PIPE if stdout else subprocess.DEVNULL
        stderr_stream = subprocess.PIPE if stderr else subprocess.DEVNULL

    # Run the command with no buffering and capturing output if we
    # want it - this will raise OSError if there was a problem running
    # the command.

    cmd = subprocess.Popen(command,
                           bufsize=0,
                           stdout=stdout_stream,
                           stderr=stderr_stream,
                           text=True,
                           errors='ignore',
                           encoding='ascii')

    # Create threads to capture stderr and/or stdout if necessary

    if stdout_stream == subprocess.PIPE:
        stdout_thread = threading.Thread(target=capture_output, args=(cmd, cmd.stdout, stdout), daemon=True)
        stdout_thread.start()
    else:
        stdout_thread = None

    if stderr_stream == subprocess.PIPE:
        stderr_thread = threading.Thread(target=capture_output, args=(cmd, cmd.stderr, stderr), daemon=True)
        stderr_thread.start()
    else:
        stderr_thread = None

    # Wait until the command terminates (and set the returncode)

    cmd.wait()

    if stdout_thread:
        stdout_thread.join()

    if stderr_thread:
        stderr_thread.join()

    # If the command failed, raise an exception

    if cmd.returncode:
        raise RunError('\n'.join(stderr_data) if stderr_data else 'Error %d running "%s"' % (cmd.returncode, ' '.join(command)),
                       status=cmd.returncode)

    # Return status, stdout, stderr (the latter 2 may be empty if we did not capture data).

    return {'status': cmd.returncode, 'stdout': stdout_data, 'stderr': stderr_data}

################################################################################

def run(command,
        stdout=None,
        stderr=None,
        output=None):
    """ Simple interface to the _process() function
        Has the same parameters, with the same defaults.
        The return value is either the data output to stdout, if any
        or the data output to stderr otherwise.
        The status code is not returned, but the function will raise an exception
        if it is non-zero """

    result = _process(command=command,
                      stdout=stdout,
                      stderr=stderr,
                      output=output)

    return result['stdout'] if result['stdout'] else result['stderr']

################################################################################

if __name__ == '__main__':
    def test_run(cmd,
                 stdout=None, stderr=None):
        """ Test wrapper for the process() function. """

        print('-' * 80)
        print('Running: %s' % (cmd if isinstance(cmd, str) else ' '.join(cmd)))

        result = _process(cmd, stdout=stdout, stderr=stderr)

        print('Status: %d' % result['status'])

    def test():
        """ Test code """

        test_run('echo nothing')

        test_run(['ls', '-l', 'run_jed'])
        test_run(['ls -l run_*'], stdout=True)
        test_run('false')
        test_run('true', stdout=sys.stdout)
        test_run(['git', 'status'], stdout=sys.stdout, stderr=sys.stderr)

        test_run(['make'], stderr=sys.stderr)
        test_run(['make'], stdout=sys.stdout, stderr=[sys.stderr])
        test_run(['make'], stdout=True)
        test_run(['make'], stdout=sys.stdout)
        test_run(['make'])

        output = []
        test_run('ls -l x*; sleep 1; echo "Bye!"', stderr=[sys.stderr, output], stdout=sys.stdout)
        print('Output=%s' % output)

    test()



# private utility functions for pydircolors
#
# Copyright 2019 Allen Wild <allenwild93@gmail.com>
# SPDX-License-Identifier: Apache-2.0

""" private/internal utility functions for pydircolors """

import os

__all__ = ['stat_at', 'readlink_at']

def stat_at(file, cwd=None, follow_symlinks=False):
    """ helper function to call os.stat on a file relative to a given directory.
    cwd should be a string, and will be opened as read-only (then closed), or an integer
    for an already-open directory file descriptor (which won't be closed).
    os.open or os.stat may raise various errors, which are passed on. """
    if isinstance(cwd, str):
        dirfd = os.open(cwd, os.O_RDONLY)
        need_to_close = True
    elif cwd is None or isinstance(cwd, int):
        dirfd = cwd
        need_to_close = False
    else:
        raise ValueError('cwd must be str, int, or None')

    try:
        return os.stat(file, dir_fd=dirfd, follow_symlinks=follow_symlinks)
    finally:
        if need_to_close:
            os.close(dirfd)

def readlink_at(file, cwd=None):
    """ helper function to call os.readlink on a file relative to a given directory.
    cwd should be a string, and will be opened as read-only (then closed), or an integer
    for an already-open directory file descriptor (which won't be closed).
    os.open or os.readlink may raise various errors, which are passed on. """
    if isinstance(cwd, str):
        dirfd = os.open(cwd, os.O_RDONLY)
        need_to_close = True
    elif cwd is None or isinstance(cwd, int):
        dirfd = cwd
        need_to_close = False
    else:
        raise ValueError('cwd must be str, int, or None')

    try:
        return os.readlink(file, dir_fd=dirfd)
    finally:
        if need_to_close:
            os.close(dirfd)



from types import SimpleNamespace

import skilleter_modules.popup as popup


class FakePanel:
    def __init__(self):
        self.update_calls = 0
        self.new_calls = []

    def new_panel(self, window):
        self.new_calls.append(window)
        return SimpleNamespace(top=lambda: None)

    def update_panels(self):
        self.update_calls += 1


class FakeWindow:
    def __init__(self):
        self.add_calls = []
        self.bkgd_calls = []

    def bkgd(self, ch, colour):
        self.bkgd_calls.append((ch, colour))

    def addstr(self, y_pos, x_pos, line, colour):
        self.add_calls.append((y_pos, x_pos, line, colour))


class FakeScreen:
    def __init__(self, keys):
        self.keys = list(keys)
        self.getch_calls = 0
        self.refresh_calls = 0

    def getmaxyx(self):
        return (24, 80)

    def refresh(self):
        self.refresh_calls += 1

    def getch(self):
        self.getch_calls += 1
        return self.keys.pop(0)


def test_waitkey_occurs_on_exit(monkeypatch):
    fake_panel = FakePanel()

    fake_curses = SimpleNamespace(
        KEY_RESIZE=999,
        panel=fake_panel,
    )

    fake_curses.color_pair = lambda colour: f'colour-{colour}'
    fake_curses.newwin = lambda h, w, y, x: FakeWindow()

    monkeypatch.setattr(popup, 'curses', fake_curses)

    keys = [fake_curses.KEY_RESIZE, ord('a')]
    screen = FakeScreen(keys)

    with popup.PopUp(screen, 'hello', 3, waitkey=True):
        enter_updates = fake_panel.update_calls
        assert screen.getch_calls == 0

    assert fake_panel.update_calls == enter_updates + 1
    assert screen.getch_calls == 2
    assert screen.refresh_calls >= 1
